#include "network_vol_private.h"


typedef struct network_vol_server_info_t {
    char *name;
    na_class_t *network_class;
    na_addr_t addr;
    _Bool connected;
    uint64_t fapl_ref;
    uint64_t obj_ref;
} network_vol_server_info_t;

typedef struct network_vol_info_t {
    network_vol_server_info_t *server_info;
} network_vol_info_t;

typedef struct network_vol_obj_common_t {
    int64_t obj_id;
    network_vol_server_info_t *server_info;
} network_vol_obj_common_t;

typedef struct network_vol_file_t {
    network_vol_obj_common_t common;    /* Must be first */
    hid_t fcpl_id;
    hid_t fapl_id;
} network_vol_file_t;

typedef struct network_vol_group_t {
    network_vol_obj_common_t common;    /* Must be first */
    hid_t gcpl_id;
    hid_t gapl_id;
} network_vol_group_t;

typedef struct network_vol_dset_t {
    network_vol_obj_common_t common;    /* Must be first */
    hid_t type_id;
    hid_t space_id;
    hid_t dcpl_id;
    hid_t dapl_id;
} network_vol_dset_t;

typedef struct network_vol_scatter_ud_t {
    void *src_buf;
    size_t src_buf_bytes_used;
} network_vol_scatter_ud_t;


static int network_vol_init(na_class_t *network_class);
static void network_vol_term(void);
static int network_vol_connect(network_vol_server_info_t *server_info);
static int network_vol_decr_obj_ref(network_vol_server_info_t *server_info);
static void *network_vol_fapl_copy(const void *_old_info);
static herr_t network_vol_fapl_free(void *info);
static void *network_vol_file_create(const char *name, unsigned flags,
    hid_t fcpl_id, hid_t fapl_id, hid_t dxpl_id, void **req);
static void *network_vol_file_open(const char *name, unsigned flags,
    hid_t fapl_id, hid_t dxpl_id, void **req);
static herr_t network_vol_file_close(void *_file, hid_t dxpl_id, void **req);
static void *network_vol_group_create(void *_obj, H5VL_loc_params_t loc_params,
    const char *name, hid_t gcpl_id, hid_t gapl_id, hid_t dxpl_id, void **req);
static void *network_vol_group_open(void *_obj, H5VL_loc_params_t loc_params,
    const char *name, hid_t gapl_id, hid_t dxpl_id, void **req);
static herr_t network_vol_group_close(void *_grp, hid_t dxpl_id, void **req);
static void *network_vol_dataset_create(void *_obj, H5VL_loc_params_t loc_params,
    const char *name, hid_t dcpl_id, hid_t dapl_id, hid_t dxpl_id, void **req);
static void *network_vol_dataset_open(void *_obj, H5VL_loc_params_t loc_params,
    const char *name, hid_t dapl_id, hid_t dxpl_id, void **req);
static herr_t network_vol_dataset_write(void *_dset, hid_t mem_type_id,
    hid_t mem_space_id, hid_t file_space_id, hid_t dxpl_id, const void *app_buf,
    void **req);
static herr_t network_vol_dataset_read(void *_dset, hid_t mem_type_id,
    hid_t mem_space_id, hid_t file_space_id, hid_t dxpl_id, void *app_buf,
    void **req);
static herr_t network_vol_dataset_close(void *_grp, hid_t dxpl_id, void **req);


/* function shipper IDs for different routines */
static hg_id_t NVOL_FILE_CREATE_ID;
static hg_id_t NVOL_FILE_OPEN_ID;
/*static hg_id_t NVOL_FILE_FLUSH_ID;*/
static hg_id_t NVOL_FILE_CLOSE_ID;
/*static hg_id_t NVOL_ATTR_CREATE_ID;
static hg_id_t NVOL_ATTR_OPEN_ID;
static hg_id_t NVOL_ATTR_READ_ID;
static hg_id_t NVOL_ATTR_WRITE_ID;
static hg_id_t NVOL_ATTR_EXISTS_ID;
static hg_id_t NVOL_ATTR_RENAME_ID;
static hg_id_t NVOL_ATTR_REMOVE_ID;
static hg_id_t NVOL_ATTR_CLOSE_ID;*/
static hg_id_t NVOL_GROUP_CREATE_ID;
static hg_id_t NVOL_GROUP_OPEN_ID;
static hg_id_t NVOL_GROUP_CLOSE_ID;
static hg_id_t NVOL_DSET_CREATE_ID;
static hg_id_t NVOL_DSET_OPEN_ID;
static hg_id_t NVOL_DSET_READ_ID;
static hg_id_t NVOL_DSET_WRITE_ID;
/*static hg_id_t NVOL_DSET_SET_EXTENT_ID;*/
static hg_id_t NVOL_DSET_CLOSE_ID;
/*static hg_id_t NVOL_DTYPE_COMMIT_ID;
static hg_id_t NVOL_DTYPE_OPEN_ID;
static hg_id_t NVOL_DTYPE_CLOSE_ID;
static hg_id_t NVOL_LINK_CREATE_ID;
static hg_id_t NVOL_LINK_MOVE_ID;
static hg_id_t NVOL_LINK_ITERATE_ID;
static hg_id_t NVOL_LINK_EXISTS_ID;
static hg_id_t NVOL_LINK_REMOVE_ID;
static hg_id_t NVOL_OBJECT_OPEN_ID;
static hg_id_t NVOL_OBJECT_COPY_ID;
static hg_id_t NVOL_OBJECT_EXISTS_ID;
static hg_id_t NVOL_OBJECT_VISIT_ID;
static hg_id_t NVOL_OBJECT_SET_COMMENT_ID;
static hg_id_t NVOL_OBJECT_GET_COMMENT_ID;
static hg_id_t NVOL_CANCEL_OP_ID;*/

static int network_vol_init_g = 0;

static hid_t network_vol_id_g = -1;

static const H5VL_class_t network_vol_g = {
    DUMMY,                                      /* class */
    "socket",                                   /* name */
    NULL,                                       /* initialize */
    NULL,                                       /* terminate */
    sizeof(network_vol_info_t),                  /* fapl_size */
    network_vol_fapl_copy,                       /* fapl_copy */
    network_vol_fapl_free,                       /* fapl_free */
    {                                           /* attribute_cls */
        NULL, //network_vol_attr_create,                /* create */
        NULL, //network_vol_attr_open,                  /* open */
        NULL, //network_vol_attr_read,                  /* read */
        NULL, //network_vol_attr_write,                 /* write */
        NULL, //network_vol_attr_get,                   /* get */
        NULL, //network_vol_attr_remove,                /* remove */
        NULL, //network_vol_attr_close                  /* close */
    },
    {                                           /* datatype_cls */
        NULL, //network_vol_datatype_commit,            /* commit */
        NULL, //network_vol_datatype_open,              /* open */
        NULL, //network_vol_datatype_get_binary,        /* get_size */
        NULL, //network_vol_datatype_close              /* close */
    },
    {                                           /* dataset_cls */
        network_vol_dataset_create,                     /* create */
        network_vol_dataset_open,                       /* open */
        network_vol_dataset_read,                       /* read */
        network_vol_dataset_write,                      /* write */
        NULL, //network_vol_dataset_set_extent,         /* set extent */
        NULL, //network_vol_dataset_get,                /* get */
        network_vol_dataset_close                       /* close */
    },
    {                                           /* file_cls */
        network_vol_file_create,                        /* create */
        network_vol_file_open,                          /* open */
        NULL, //network_vol_file_flush,                 /* flush */
        NULL, //network_vol_file_get,                   /* get */
        NULL, //network_vol_file_misc,                  /* misc */
        NULL, //network_vol_file_optional,              /* optional */
        network_vol_file_close                  /* close */
    },
    {                                          /* group_cls */
        network_vol_group_create,                       /* create */
        network_vol_group_open,                         /* open */
        NULL, //network_vol_group_get,                  /* get */
        network_vol_group_close                 /* close */
    },
    {                                          /* link_cls */
        NULL, //network_vol_link_create,                /* create */
        NULL, //network_vol_link_move,                  /* move */
        NULL, //network_vol_link_iterate,               /* iterate */
        NULL, //network_vol_link_get,                   /* get */
        NULL, //network_vol_link_remove                 /* remove */
    },
    {                                          /* object_cls */
        NULL, //network_vol_object_open,                /* open */
        NULL, //network_vol_object_copy,                /* copy */
        NULL, //network_vol_object_visit,               /* visit */
        NULL, //network_vol_object_get,                 /* get */
        NULL, //network_vol_object_misc,                /* misc */
        NULL, //network_vol_object_optional,            /* optional */
        NULL, //network_vol_object_close                /* close */
    }
};


static int
network_vol_init(na_class_t *network_class)
{
    if(HG_Init(network_class) != HG_SUCCESS)
        ERROR("Could not initialize function shipper");

    if(HG_Bulk_init(network_class) != HG_SUCCESS)
        ERROR("Could not initialize bulk data shipper");

    if(0 != atexit(network_vol_term))
        ERROR("Could not register atexit function");

    /* Register function and encoding/decoding functions */
    NVOL_FILE_CREATE_ID = MERCURY_REGISTER("file_create", file_create_in_t, servid_t);
    NVOL_FILE_OPEN_ID   = MERCURY_REGISTER("file_open", file_open_in_t, file_open_out_t);
    //NVOL_FILE_FLUSH_ID  = MERCURY_REGISTER("file_flush", file_flush_in_t, ret_t);
    NVOL_FILE_CLOSE_ID  = MERCURY_REGISTER("file_close", servid_t, ret_t);

    /*NVOL_ATTR_CREATE_ID = MERCURY_REGISTER("attr_create", attr_create_in_t, attr_create_out_t);
    NVOL_ATTR_OPEN_ID   = MERCURY_REGISTER("attr_open", attr_open_in_t, attr_open_out_t);
    NVOL_ATTR_READ_ID   = MERCURY_REGISTER("attr_read", attr_io_in_t, ret_t);
    NVOL_ATTR_WRITE_ID  = MERCURY_REGISTER("attr_write", attr_io_in_t, ret_t);
    NVOL_ATTR_EXISTS_ID = MERCURY_REGISTER("attr_exists", attr_op_in_t, htri_t);
    NVOL_ATTR_REMOVE_ID = MERCURY_REGISTER("attr_remove", attr_op_in_t, ret_t);
    NVOL_ATTR_CLOSE_ID  = MERCURY_REGISTER("attr_close", attr_close_in_t, ret_t);*/

    NVOL_GROUP_CREATE_ID = MERCURY_REGISTER("group_create", group_create_in_t, servid_t);
    NVOL_GROUP_OPEN_ID   = MERCURY_REGISTER("group_open", group_open_in_t, group_open_out_t);
    NVOL_GROUP_CLOSE_ID  = MERCURY_REGISTER("group_close", servid_t, ret_t);

    NVOL_DSET_CREATE_ID = MERCURY_REGISTER("dset_create", dset_create_in_t, servid_t);
    NVOL_DSET_OPEN_ID   = MERCURY_REGISTER("dset_open", dset_open_in_t, dset_open_out_t);
    NVOL_DSET_READ_ID   = MERCURY_REGISTER("dset_read", dset_io_in_t, ret_t);
    NVOL_DSET_WRITE_ID  = MERCURY_REGISTER("dset_write", dset_io_in_t, ret_t);
    /*NVOL_DSET_SET_EXTENT_ID = MERCURY_REGISTER("dset_set_extent", 
                                               dset_set_extent_in_t, ret_t);*/
    NVOL_DSET_CLOSE_ID  = MERCURY_REGISTER("dset_close", servid_t, ret_t);

    /*NVOL_DTYPE_COMMIT_ID = MERCURY_REGISTER("dtype_commit", dtype_commit_in_t, dtype_commit_out_t);
    NVOL_DTYPE_OPEN_ID   = MERCURY_REGISTER("dtype_open", dtype_open_in_t, dtype_open_out_t);
    NVOL_DTYPE_CLOSE_ID  = MERCURY_REGISTER("dtype_close", dtype_close_in_t, ret_t);*/

    if((network_vol_id_g = H5VLregister(&network_vol_g)) < 0)
        ERROR("H5VLregister failed");

    network_vol_init_g = 1;

    return(SUCCEED);

error:
    return(FAIL);
} /* end network_vol_init() */


static void
network_vol_term(void)
{
    /* Call H5close(), so we are sure HDF5 won't reenter this library after we
     * finalize mercury */
    (void)H5close();

    (void)HG_Bulk_finalize();
    (void)HG_Finalize();

    return;
} /* end network_vol_term() */


int
hdf5_network_vol_set(hid_t fapl_id, na_class_t *network_class,
    char *server_name)
{
    network_vol_info_t vol_info = {NULL};

    if(!network_class)
        ERROR("no network class provided");
    if(!server_name)
        ERROR("no server provided");

    if(!network_vol_init_g)
        if(network_vol_init(network_class) < 0)
            ERROR("network_vol_init failed");

    if(NULL == (vol_info.server_info = (network_vol_server_info_t *)malloc(sizeof(network_vol_server_info_t))))
        ERROR("calloc of server info failed");

    /* Copy name */
    if(NULL == (vol_info.server_info->name = strdup(server_name)))
        ERROR("strdup of server name failed");

    /* Copy network class */
    vol_info.server_info->network_class = network_class;

    if(H5Pset_vol(fapl_id, network_vol_id_g, &vol_info) < 0)
        ERROR("H5Pset_vol failed");

    vol_info.server_info->connected = FALSE;
    vol_info.server_info->fapl_ref = 1;
    vol_info.server_info->obj_ref = 0;

    return(SUCCEED);

error:
    if(vol_info.server_info)
        free(vol_info.server_info);

    return(FAIL);
} /* end network_vol_set() */


static int
network_vol_connect(network_vol_server_info_t *server_info)
{
    assert(server_info->obj_ref == 0);

    /* Look up server address (this opens the connection) */
    if(NA_Addr_lookup(server_info->network_class, server_info->name, &server_info->addr) != NA_SUCCESS)
        ERROR("Could not find server addresss\n");
    server_info->connected = TRUE;

    return(SUCCEED);

error:
    return(FAIL);
} /* end network_vol_connect() */


static void
network_vol_server_info_free(network_vol_server_info_t *server_info)
{
    free(server_info->name);
    free(server_info);

    return;
} /* end network_vol_server_info_free() */


static int
network_vol_decr_obj_ref(network_vol_server_info_t *server_info)
{
    assert(server_info->obj_ref > 0);

    /* Decrement reference on server */
    server_info->obj_ref--;
    if(server_info->obj_ref == 0) {
        /* We closed the last object holding this connection open, free address
         * (this closes the connection) */
        if(NA_Addr_free(server_info->network_class, server_info->addr) != NA_SUCCESS)
            ERROR("Could not free addr\n");
        server_info->connected = FALSE;

        /* If this is also the last item holding the server_info struct open,
         * free it */
        if(server_info->fapl_ref == 0)
            network_vol_server_info_free(server_info);
    } /* end if */

    return(SUCCEED);

error:
    return(FAIL);
} /* end network_vol_obj_free() */


static void *
network_vol_fapl_copy(const void *_old_info)
{
    const network_vol_info_t *old_info = (const network_vol_info_t *)_old_info;
    network_vol_info_t *new_info = NULL;

    /* Copy top level struct */
    if(NULL == (new_info = (network_vol_info_t *)malloc(sizeof(network_vol_info_t))))
        ERROR("malloc of vol info failed");
    memcpy(new_info, old_info, sizeof(network_vol_info_t));

    new_info->server_info->fapl_ref++;

    return(new_info);

error:
    if(new_info)
        free(new_info);

    return(NULL);
} /* end network_vol_fapl_copy() */


static herr_t
network_vol_fapl_free(void *_info)
{
    network_vol_info_t *info = (network_vol_info_t *)_info;

    info->server_info->fapl_ref--;

    if(info->server_info->fapl_ref == 0 && info->server_info->obj_ref == 0) {
        network_vol_server_info_free(info->server_info);
    } /* end if */

    free(info);

    return(SUCCEED);
} /* end network_vol_fapl_free() */


static void *
network_vol_file_create(const char *name, unsigned flags, hid_t fcpl_id,
    hid_t fapl_id, hid_t dxpl_id, void **req)
{
    hg_request_t hg_req;        /* Local function shipper request */
    hg_status_t hg_status;
    file_create_in_t input;
    const network_vol_info_t *vol_info;
    network_vol_file_t *file = NULL;

    /* Retrieve vol info and allocate file object */
    if(NULL == (vol_info = (const network_vol_info_t *)H5Pget_vol_info(fapl_id)))
        ERROR("H5Pget_vol_info failed");
    if(NULL == (file = (network_vol_file_t *)malloc(sizeof(network_vol_file_t))))
        ERROR("malloc of file object failed");

    /* Set the input structure for the HG encode routine */
    input.name = name;
    input.flags = flags;
    input.fcpl_id = fcpl_id;
    input.fapl_id = fapl_id;

    /* Connect to the server, if necessary */
    if(vol_info->server_info->obj_ref == 0)
        if(network_vol_connect(vol_info->server_info) < 0)
            ERROR("network_vol_connect failed");

    /* Forward the call to the function shipper */
    if(HG_Forward(vol_info->server_info->addr, NVOL_FILE_CREATE_ID, &input, &file->common.obj_id, &hg_req) != HG_SUCCESS)
        ERROR("failed to ship file create");

    /* Wait for call to complete */
    if(HG_Wait(hg_req, NVOL_MAX_IDLE_TIME, &hg_status) != HG_SUCCESS)
        ERROR("error during wait");
    if(!hg_status)
        ERROR("operation did not complete");

    /* Check if the HDF5 call failed */
    if(file->common.obj_id < 0)
        ERROR("operation on server failed");

    /* Set up file object */
    vol_info->server_info->obj_ref++;
    file->common.server_info = vol_info->server_info;
    if((file->fcpl_id = H5Pcopy(fcpl_id)) < 0)
        ERROR("failed to copy fcpl");
    if((file->fapl_id = H5Pcopy(fapl_id)) < 0)
        ERROR("failed to copy fapl");

    return(file);

error:
    if(file)
        free(file);

    return(NULL);
} /* end network_vol_file_create() */


static void *
network_vol_file_open(const char *name, unsigned flags, hid_t fapl_id, hid_t
    dxpl_id, void **req)
{
    hg_request_t hg_req;        /* Local function shipper request */
    hg_status_t hg_status;
    file_open_in_t input;
    file_open_out_t output;
    const network_vol_info_t *vol_info;
    network_vol_file_t *file = NULL;

    /* Retrieve vol info and allocate file object */
    if(NULL == (vol_info = (const network_vol_info_t *)H5Pget_vol_info(fapl_id)))
        ERROR("H5Pget_vol_info failed");
    if(NULL == (file = (network_vol_file_t *)malloc(sizeof(network_vol_file_t))))
        ERROR("malloc of file object failed");

    /* Set the input structure for the HG encode routine */
    input.name = name;
    input.flags = flags;
    input.fapl_id = fapl_id;

    /* Connect to the server, if necessary */
    if(vol_info->server_info->obj_ref == 0)
        if(network_vol_connect(vol_info->server_info) < 0)
            ERROR("network_vol_connect failed");

    /* Forward the call to the function shipper */
    if(HG_Forward(vol_info->server_info->addr, NVOL_FILE_OPEN_ID, &input, &output, &hg_req) != HG_SUCCESS)
        ERROR("failed to ship file open");

    /* Wait for call to complete */
    if(HG_Wait(hg_req, NVOL_MAX_IDLE_TIME, &hg_status) != HG_SUCCESS)
        ERROR("error during wait");
    if(!hg_status)
        ERROR("operation did not complete");

    /* Check if the HDF5 call failed */
    if(output.file_id < 0)
        ERROR("operation on server failed");

    /* Set up file object */
    file->common.obj_id = output.file_id;
    vol_info->server_info->obj_ref++;
    file->common.server_info = vol_info->server_info;
    file->fcpl_id = output.fcpl_id;
    if((file->fapl_id = H5Pcopy(fapl_id)) < 0)
        ERROR("failed to copy fapl");

    /* Increment reference counts on property lists decoded from server so they
     * aren't closed when the output is freed by mercury */
    if(H5Iinc_ref(output.fcpl_id) < 0)
        ERROR("failed to increment reference count");

    return(file);

error:
    if(file)
        free(file);

    return(NULL);
} /* end network_vol_file_open() */


static herr_t
network_vol_file_close(void *_file, hid_t dxpl_id, void **req)
{
    network_vol_file_t *file = (network_vol_file_t *)_file;
    hg_request_t hg_req;        /* Local function shipper request */
    hg_status_t hg_status;
    int status;

    assert(file->common.server_info);
    assert(file->common.server_info->obj_ref > 0);

    /* Forward the call to the function shipper */
    if(HG_Forward(file->common.server_info->addr, NVOL_FILE_CLOSE_ID, &file->common.obj_id, &status, &hg_req) != HG_SUCCESS)
        ERROR("failed to ship file close");

    /* Wait for call to complete */
    if(HG_Wait(hg_req, NVOL_MAX_IDLE_TIME, &hg_status) != HG_SUCCESS)
        ERROR("error during wait");
    if(!hg_status)
        ERROR("operation did not complete");

    /* Check if the HDF5 call failed */
    if(status < 0)
        ERROR("operation on server failed");

    /* Free the file object and associated property lists */
    if(network_vol_decr_obj_ref(file->common.server_info) < 0)
        ERROR("network_vol_decr_obj_ref failed");
    if(H5Pclose(file->fcpl_id) < 0)
        ERROR("failed to close fcpl");
    if(H5Pclose(file->fapl_id) < 0)
        ERROR("failed to close fapl");
    free(file);
    
    file = NULL;

    return(SUCCEED);

error:
    return(FAIL);
} /* end network_vol_file_close() */


static void *
network_vol_group_create(void *_obj, H5VL_loc_params_t loc_params,
    const char *name, hid_t gcpl_id, hid_t gapl_id, hid_t dxpl_id, void **req)
{
    network_vol_obj_common_t *obj = (network_vol_obj_common_t *)_obj;
    hid_t lcpl_id = -1;
    hg_request_t hg_req;        /* Local function shipper request */
    hg_status_t hg_status;
    group_create_in_t input;
    network_vol_group_t *grp = NULL;

    assert(loc_params.type == H5VL_OBJECT_BY_SELF);
    assert(obj->server_info);
    assert(obj->server_info->obj_ref > 0);

    /* Allocate group object */
    if(NULL == (grp = (network_vol_group_t *)malloc(sizeof(network_vol_group_t))))
        ERROR("malloc of group object failed");

    /* Get lcpl id from gcpl */
    if(H5Pget(gcpl_id, H5VL_GRP_LCPL_ID, &lcpl_id) < 0)
        ERROR("H5Pget failed to get lcpl id from gcpl");

    /* Set the input structure for the HG encode routine */
    input.loc_id = obj->obj_id;
    input.name = name;
    input.gapl_id = gapl_id;
    input.gcpl_id = gcpl_id;
    input.lcpl_id = lcpl_id;

    /* Forward the call to the function shipper */
    if(HG_Forward(obj->server_info->addr, NVOL_GROUP_CREATE_ID, &input, &grp->common.obj_id, &hg_req) != HG_SUCCESS)
        ERROR("failed to ship group create");

    /* Wait for call to complete */
    if(HG_Wait(hg_req, NVOL_MAX_IDLE_TIME, &hg_status) != HG_SUCCESS)
        ERROR("error during wait");
    if(!hg_status)
        ERROR("operation did not complete");

    /* Check if the HDF5 call failed */
    if(grp->common.obj_id < 0)
        ERROR("operation on server failed");

    /* Set up group object */
    obj->server_info->obj_ref++;
    grp->common.server_info = obj->server_info;
    if((grp->gcpl_id = H5Pcopy(gcpl_id)) < 0)
        ERROR("failed to copy gcpl");
    if((grp->gapl_id = H5Pcopy(gapl_id)) < 0)
        ERROR("failed to copy gapl");

    return(grp);

error:
    if(grp)
        free(grp);

    return(NULL);
} /* end network_vol_group_create() */


static void *
network_vol_group_open(void *_obj, H5VL_loc_params_t loc_params,
    const char *name, hid_t gapl_id, hid_t dxpl_id, void **req)
{
    network_vol_obj_common_t *obj = (network_vol_obj_common_t *)_obj;
    hg_request_t hg_req;        /* Local function shipper request */
    hg_status_t hg_status;
    group_open_in_t input;
    group_open_out_t output;
    network_vol_group_t *grp = NULL;

    assert(loc_params.type == H5VL_OBJECT_BY_SELF);
    assert(obj->server_info);
    assert(obj->server_info->obj_ref > 0);

    /* Allocate group object */
    if(NULL == (grp = (network_vol_group_t *)malloc(sizeof(network_vol_group_t))))
        ERROR("malloc of group object failed");

    /* Set the input structure for the HG encode routine */
    input.loc_id = obj->obj_id;
    input.name = name;
    input.gapl_id = gapl_id;

    /* Forward the call to the function shipper */
    if(HG_Forward(obj->server_info->addr, NVOL_GROUP_OPEN_ID, &input, &output, &hg_req) != HG_SUCCESS)
        ERROR("failed to ship group open");

    /* Wait for call to complete */
    if(HG_Wait(hg_req, NVOL_MAX_IDLE_TIME, &hg_status) != HG_SUCCESS)
        ERROR("error during wait");
    if(!hg_status)
        ERROR("operation did not complete");

    /* Check if the HDF5 call failed */
    if(output.grp_id < 0)
        ERROR("operation on server failed");

    /* Set up group object */
    grp->common.obj_id = output.grp_id;
    obj->server_info->obj_ref++;
    grp->common.server_info = obj->server_info;
    grp->gcpl_id = output.gcpl_id;
    if((grp->gapl_id = H5Pcopy(gapl_id)) < 0)
        ERROR("failed to copy gapl");

    /* Increment reference counts on property lists decoded from server so they
     * aren't closed when the output is freed by mercury */
    if(H5Iinc_ref(output.gcpl_id) < 0)
        ERROR("failed to increment reference count");

    return(grp);

error:
    if(grp)
        free(grp);

    return(NULL);
} /* end network_vol_group_open() */


static herr_t
network_vol_group_close(void *_grp, hid_t dxpl_id, void **req)
{
    network_vol_group_t *grp = (network_vol_group_t *)_grp;
    hg_request_t hg_req;        /* Local function shipper request */
    hg_status_t hg_status;
    int status;

    assert(grp->common.server_info);
    assert(grp->common.server_info->obj_ref > 0);

    /* Forward the call to the function shipper */
    if(HG_Forward(grp->common.server_info->addr, NVOL_GROUP_CLOSE_ID, &grp->common.obj_id, &status, &hg_req) != HG_SUCCESS)
        ERROR("failed to ship group close");

    /* Wait for call to complete */
    if(HG_Wait(hg_req, NVOL_MAX_IDLE_TIME, &hg_status) != HG_SUCCESS)
        ERROR("error during wait");
    if(!hg_status)
        ERROR("operation did not complete");

    /* Check if the HDF5 call failed */
    if(status < 0)
        ERROR("operation on server failed");

    /* Free the group object and associated property lists */
    if(network_vol_decr_obj_ref(grp->common.server_info) < 0)
        ERROR("network_vol_decr_obj_ref failed");
    if(H5Pclose(grp->gcpl_id) < 0)
        ERROR("failed to close gcpl");
    if(H5Pclose(grp->gapl_id) < 0)
        ERROR("failed to close gapl");
    free(grp);
    
    grp = NULL;

    return(SUCCEED);

error:
    return(FAIL);
} /* end network_vol_group_close() */


static void *
network_vol_dataset_create(void *_obj, H5VL_loc_params_t loc_params,
    const char *name, hid_t dcpl_id, hid_t dapl_id, hid_t dxpl_id, void **req)
{
    network_vol_obj_common_t *obj = (network_vol_obj_common_t *)_obj;
    hid_t type_id = -1;
    hid_t space_id = -1;
    hid_t lcpl_id = -1;
    hg_request_t hg_req;        /* Local function shipper request */
    hg_status_t hg_status;
    dset_create_in_t input;
    network_vol_dset_t *dset = NULL;

    assert(loc_params.type == H5VL_OBJECT_BY_SELF);
    assert(obj->server_info);
    assert(obj->server_info->obj_ref > 0);

    /* Allocate dataset object */
    if(NULL == (dset = (network_vol_dset_t *)malloc(sizeof(network_vol_dset_t))))
        ERROR("malloc of group object failed");

    /*
     * Get dtype, space, and lcpl ids from dcpl
     */
    if(H5Pget(dcpl_id, H5VL_DSET_TYPE_ID, &type_id) < 0)
        ERROR("H5Pget failed to get datatype id from dcpl");
    if(H5Pget(dcpl_id, H5VL_DSET_SPACE_ID, &space_id) < 0)
        ERROR("H5Pget failed to get dataspace id from dcpl");
    if(H5Pget(dcpl_id, H5VL_DSET_LCPL_ID, &lcpl_id) < 0)
        ERROR("H5Pget failed to get lcpl id from dcpl");

    /* Set the input structure for the HG encode routine */
    input.loc_id = obj->obj_id;
    input.name = name;
    input.dapl_id = dapl_id;
    input.dcpl_id = dcpl_id;
    input.lcpl_id = lcpl_id;
    input.type_id = type_id;
    input.space_id = space_id;

    /* Forward the call to the function shipper */
    if(HG_Forward(obj->server_info->addr, NVOL_DSET_CREATE_ID, &input, &dset->common.obj_id, &hg_req) != HG_SUCCESS)
        ERROR("failed to ship dataset create");

    /* Wait for call to complete */
    if(HG_Wait(hg_req, NVOL_MAX_IDLE_TIME, &hg_status) != HG_SUCCESS)
        ERROR("error during wait");
    if(!hg_status)
        ERROR("operation did not complete");

    /* Check if the HDF5 call failed */
    if(dset->common.obj_id < 0)
        ERROR("operation on server failed");

    /* Set up dataset object */
    obj->server_info->obj_ref++;
    dset->common.server_info = obj->server_info;
    if((dset->type_id = H5Tcopy(type_id)) < 0)
        ERROR("failed to copy datatype");
    if((dset->space_id = H5Scopy(space_id)) < 0)
        ERROR("failed to copy dataspace");
    if((dset->dcpl_id = H5Pcopy(dcpl_id)) < 0)
        ERROR("failed to copy dcpl");
    if((dset->dapl_id = H5Pcopy(dapl_id)) < 0)
        ERROR("failed to copy dapl");

    return(dset);

error:
    if(dset)
        free(dset);

    return(NULL);
} /* end network_vol_dset_create() */


static void *
network_vol_dataset_open(void *_obj, H5VL_loc_params_t loc_params,
    const char *name, hid_t dapl_id, hid_t dxpl_id, void **req)
{
    network_vol_obj_common_t *obj = (network_vol_obj_common_t *)_obj;
    hg_request_t hg_req;        /* Local function shipper request */
    hg_status_t hg_status;
    dset_open_in_t input;
    dset_open_out_t output;
    network_vol_dset_t *dset = NULL;

    assert(loc_params.type == H5VL_OBJECT_BY_SELF);
    assert(obj->server_info);
    assert(obj->server_info->obj_ref > 0);

    /* Allocate dataset object */
    if(NULL == (dset = (network_vol_dset_t *)malloc(sizeof(network_vol_dset_t))))
        ERROR("malloc of group object failed");

    /* Set the input structure for the HG encode routine */
    input.loc_id = obj->obj_id;
    input.name = name;
    input.dapl_id = dapl_id;

    /* Forward the call to the function shipper */
    if(HG_Forward(obj->server_info->addr, NVOL_DSET_OPEN_ID, &input, &output, &hg_req) != HG_SUCCESS)
        ERROR("failed to ship dataset open");

    /* Wait for call to complete */
    if(HG_Wait(hg_req, NVOL_MAX_IDLE_TIME, &hg_status) != HG_SUCCESS)
        ERROR("error during wait");
    if(!hg_status)
        ERROR("operation did not complete");

    /* Check if the HDF5 call failed */
    if(output.dset_id < 0)
        ERROR("operation on server failed");

    /* Set up dataset object */
    dset->common.obj_id = output.dset_id;
    obj->server_info->obj_ref++;
    dset->common.server_info = obj->server_info;
    dset->type_id = output.type_id;
    dset->space_id = output.space_id;
    dset->dcpl_id = output.dcpl_id;
    if((dset->dapl_id = H5Pcopy(dapl_id)) < 0)
        ERROR("failed to copy dapl");

    /* Increment reference counts on ids decoded from server so they aren't
     * closed when the output is freed by mercury */
    if(H5Iinc_ref(output.type_id) < 0)
        ERROR("failed to increment reference count");
    if(H5Iinc_ref(output.space_id) < 0)
        ERROR("failed to increment reference count");
    if(H5Iinc_ref(output.dcpl_id) < 0)
        ERROR("failed to increment reference count");

    return(dset);

error:
    if(dset)
        free(dset);

    return(NULL);
} /* end network_vol_dataset_open() */


static herr_t
nvol_dataset_read_scatter_cb(void **src_buf, size_t *src_buf_bytes_used,
    void *_udata)
{
    network_vol_scatter_ud_t *udata = (network_vol_scatter_ud_t *)_udata;

    *src_buf = udata->src_buf;
    *src_buf_bytes_used = udata->src_buf_bytes_used;

    return 0;
} /* end nvol_dataset_read_scatter_cb() */


static herr_t
network_vol_dataset_read(void *_dset, hid_t mem_type_id, hid_t mem_space_id,
    hid_t file_space_id, hid_t dxpl_id, void *app_buf, void **req)
{
    network_vol_dset_t *dset = (network_vol_dset_t *)_dset;
    hg_request_t hg_req;        /* Local function shipper request */
    hg_status_t hg_status;
    dset_io_in_t input;
    void *tmp_buf = NULL;
    void *bulk_buf = NULL;
    size_t bulk_buf_len;
    hid_t tmp_space_id;
    size_t mem_type_elem_size;
    hssize_t mem_extent_npoints;
    hssize_t mem_select_npoints;
    htri_t tri_ret;
    network_vol_scatter_ud_t scatter_udata;
    int status;

    assert(dset->common.server_info);
    assert(dset->common.server_info->obj_ref > 0);

    /* Check for unsupported datatypes */
    if((tri_ret = H5Tdetect_class(mem_type_id, H5T_VLEN)) < 0)
        ERROR("H5Tdetect_class failed");
    if(tri_ret)
        ERROR("vlen types not supported");
    if((tri_ret = H5Tdetect_class(mem_type_id, H5T_REFERENCE)) < 0)
        ERROR("H5Tdetect_class failed");
    if(tri_ret)
        ERROR("reference types not supported");

    /* Calculate application buffer length */
    /* Use mem_space_id, unless it is H5S_ALL, in which case use file_space_id,
     * unless it is also H5S_ALL, in which case use dset->space_id */
    if(mem_space_id == H5S_ALL) {
        if(file_space_id == H5S_ALL)
            tmp_space_id = dset->space_id;
        else
            tmp_space_id = file_space_id;
    } /* end if */
    else
        tmp_space_id = mem_space_id;
    if(0 == (mem_type_elem_size = H5Tget_size(mem_type_id)))
        ERROR("H5Tget_size failed");
    if((mem_extent_npoints = H5Sget_simple_extent_npoints(tmp_space_id)) < 0)
        ERROR("H5Sget_simple_extent_npoints failed");
    if((mem_select_npoints = H5Sget_select_npoints(tmp_space_id)) < 0)
        ERROR("H5Sget_simple_select_npoints failed");
    assert(mem_extent_npoints >= mem_select_npoints);
    bulk_buf_len = mem_type_elem_size * (size_t)mem_select_npoints;

    /* Allocate temporary buffer if we need to scatter to app_buf */
    if(mem_extent_npoints != mem_select_npoints) {
        if(NULL == (tmp_buf = malloc(bulk_buf_len)))
            ERROR("malloc failed");
        bulk_buf = tmp_buf;
    } /* end if */
    else
        bulk_buf = app_buf;

    /* Set the input structure for the HG encode routine */
    input.dset_id = dset->common.obj_id;
    input.mem_type_id = mem_type_id;
    input.mem_space_id = mem_space_id == H5S_ALL ? NVOL_DEF_ID : mem_space_id;
    input.file_space_id = file_space_id == H5S_ALL ? NVOL_DEF_ID : file_space_id;
    input.dxpl_id = dxpl_id;

    /* Register memory */
    if(HG_SUCCESS != HG_Bulk_handle_create(bulk_buf, bulk_buf_len, HG_BULK_READWRITE, &input.bulk_handle))
        ERROR("can't create Bulk Data Handle");

    /* Forward the call to the function shipper */
    if(HG_Forward(dset->common.server_info->addr, NVOL_DSET_READ_ID, &input, &status, &hg_req) != HG_SUCCESS)
        ERROR("failed to ship dataset read");

    /* Wait for call to complete */
    if(HG_Wait(hg_req, NVOL_MAX_IDLE_TIME, &hg_status) != HG_SUCCESS)
        ERROR("error during wait");
    if(!hg_status)
        ERROR("operation did not complete");

    /* Check if the HDF5 call failed */
    if(status < 0)
        ERROR("operation on server failed");

    /* Scatter data if necessary */
    if(tmp_buf) {
        assert(tmp_buf == bulk_buf);
        scatter_udata.src_buf = bulk_buf;
        scatter_udata.src_buf_bytes_used = bulk_buf_len;
        if(H5Dscatter(nvol_dataset_read_scatter_cb, &scatter_udata, mem_type_id, tmp_space_id, app_buf) < 0)
            ERROR("failed to scatter data to app buffer");
        free(tmp_buf);
    } /* end if */

    return(SUCCEED);

error:
    if(tmp_buf)
        free(tmp_buf);
 
    return(FAIL);
} /* end network_vol_dataset_read() */


static herr_t
network_vol_dataset_write(void *_dset, hid_t mem_type_id, hid_t mem_space_id,
    hid_t file_space_id, hid_t dxpl_id, const void *app_buf, void **req)
{
    network_vol_dset_t *dset = (network_vol_dset_t *)_dset;
    hg_request_t hg_req;        /* Local function shipper request */
    hg_status_t hg_status;
    dset_io_in_t input;
    void *tmp_buf = NULL;
    void *bulk_buf = NULL;
    size_t bulk_buf_len;
    hid_t tmp_space_id;
    size_t mem_type_elem_size;
    hssize_t mem_extent_npoints;
    hssize_t mem_select_npoints;
    htri_t tri_ret;
    int status;

    assert(dset->common.server_info);
    assert(dset->common.server_info->obj_ref > 0);

    /* Check for unsupported datatypes */
    if((tri_ret = H5Tdetect_class(mem_type_id, H5T_VLEN)) < 0)
        ERROR("H5Tdetect_class failed");
    if(tri_ret)
        ERROR("vlen types not supported");
    if((tri_ret = H5Tdetect_class(mem_type_id, H5T_REFERENCE)) < 0)
        ERROR("H5Tdetect_class failed");
    if(tri_ret)
        ERROR("reference types not supported");

    /* Calculate application buffer length */
    /* Use mem_space_id, unless it is H5S_ALL, in which case use file_space_id,
     * unless it is also H5S_ALL, in which case use dset->space_id */
    if(mem_space_id == H5S_ALL) {
        if(file_space_id == H5S_ALL)
            tmp_space_id = dset->space_id;
        else
            tmp_space_id = file_space_id;
    } /* end if */
    else
        tmp_space_id = mem_space_id;
    if(0 == (mem_type_elem_size = H5Tget_size(mem_type_id)))
        ERROR("H5Tget_size failed");
    if((mem_extent_npoints = H5Sget_simple_extent_npoints(tmp_space_id)) < 0)
        ERROR("H5Sget_simple_extent_npoints failed");
    if((mem_select_npoints = H5Sget_select_npoints(tmp_space_id)) < 0)
        ERROR("H5Sget_simple_select_npoints failed");
    assert(mem_extent_npoints >= mem_select_npoints);
    bulk_buf_len = mem_type_elem_size * (size_t)mem_select_npoints;

    /* Allocate temporary buffer and gather data from app_buf, if necessary */
    if(mem_extent_npoints != mem_select_npoints) {
        if(NULL == (tmp_buf = malloc(bulk_buf_len)))
            ERROR("malloc failed");
        bulk_buf = tmp_buf;

        if(H5Dgather(tmp_space_id, app_buf, mem_type_id, bulk_buf_len, bulk_buf, NULL, NULL) < 0)
            ERROR("unable to gather data from app buffer");
    } /* end if */
    else
        bulk_buf = (void *)app_buf;

    /* Set the input structure for the HG encode routine */
    input.dset_id = dset->common.obj_id;
    input.mem_type_id = mem_type_id;
    input.mem_space_id = mem_space_id == H5S_ALL ? NVOL_DEF_ID : mem_space_id;
    input.file_space_id = file_space_id == H5S_ALL ? NVOL_DEF_ID : file_space_id;
    input.dxpl_id = dxpl_id;

    /* Register memory */
    if(HG_SUCCESS != HG_Bulk_handle_create(bulk_buf, bulk_buf_len, HG_BULK_READ_ONLY, &input.bulk_handle))
        ERROR("can't create Bulk Data Handle");

    /* Forward the call to the function shipper */
    if(HG_Forward(dset->common.server_info->addr, NVOL_DSET_WRITE_ID, &input, &status, &hg_req) != HG_SUCCESS)
        ERROR("failed to ship dataset write");

    /* Wait for call to complete */
    if(HG_Wait(hg_req, NVOL_MAX_IDLE_TIME, &hg_status) != HG_SUCCESS)
        ERROR("error during wait");
    if(!hg_status)
        ERROR("operation did not complete");

    /* Check if the HDF5 call failed */
    if(status < 0)
        ERROR("operation on server failed");

    return(SUCCEED);

error:
    return(FAIL);
} /* end network_vol_dataset_write() */


static herr_t
network_vol_dataset_close(void *_dset, hid_t dxpl_id, void **req)
{
    network_vol_dset_t *dset = (network_vol_dset_t *)_dset;
    hg_request_t hg_req;        /* Local function shipper request */
    hg_status_t hg_status;
    int status;

    assert(dset->common.server_info);
    assert(dset->common.server_info->obj_ref > 0);

    /* Forward the call to the function shipper */
    if(HG_Forward(dset->common.server_info->addr, NVOL_DSET_CLOSE_ID, &dset->common.obj_id, &status, &hg_req) != HG_SUCCESS)
        ERROR("failed to ship dataset close");

    /* Wait for call to complete */
    if(HG_Wait(hg_req, NVOL_MAX_IDLE_TIME, &hg_status) != HG_SUCCESS)
        ERROR("error during wait");
    if(!hg_status)
        ERROR("operation did not complete");

    /* Check if the HDF5 call failed */
    if(status < 0)
        ERROR("operation on server failed");

    /* Free the dataset object, associated property lists, type, and space */
    if(network_vol_decr_obj_ref(dset->common.server_info) < 0)
        ERROR("network_vol_decr_obj_ref failed");
    if(H5Tclose(dset->type_id) < 0)
        ERROR("failed to close datatype");
    if(H5Sclose(dset->space_id) < 0)
        ERROR("failed to close dataspace");
    if(H5Pclose(dset->dcpl_id) < 0)
        ERROR("failed to close dcpl");
    if(H5Pclose(dset->dapl_id) < 0)
        ERROR("failed to close dapl");
    free(dset);
    
    dset = NULL;

    return(SUCCEED);

error:
    return(FAIL);
} /* end network_vol_dataset_close() */

