#include "network_vol_private.h"

#include <axe.h>
#include <mercury_handler.h>


typedef struct obj_id_ud_t {
    int64_t obj_id;
    hg_handle_t hg_handle;
} obj_id_ud_t;

typedef struct file_create_ud_t {
    file_create_in_t input;
    hg_handle_t hg_handle;
} file_create_ud_t;
typedef struct file_open_ud_t {
    file_open_in_t input;
    hg_handle_t hg_handle;
} file_open_ud_t;

typedef struct group_create_ud_t {
    group_create_in_t input;
    hg_handle_t hg_handle;
} group_create_ud_t;
typedef struct group_open_ud_t {
    group_open_in_t input;
    hg_handle_t hg_handle;
} group_open_ud_t;

typedef struct dset_create_ud_t {
    dset_create_in_t input;
    hg_handle_t hg_handle;
} dset_create_ud_t;
typedef struct dset_open_ud_t {
    dset_open_in_t input;
    hg_handle_t hg_handle;
} dset_open_ud_t;
typedef struct dset_io_ud_t {
    dset_io_in_t input;
    hg_handle_t hg_handle;
} dset_io_ud_t;


static AXE_engine_t engine;
static _Bool nvol_shutdown = 0;


static void
nvol_file_create_handler_cb(AXE_engine_t axe_engine,
    size_t num_necessary_parents, AXE_task_t necessary_parents[],
    size_t num_sufficient_parents, AXE_task_t sufficient_parents[],
    void *_udata)
{
    file_create_ud_t *udata = (file_create_ud_t *)_udata;
    int64_t file_id;

    /* Create file.  No need to check for errors because we wouldn't do anything
     * different - just forward result to client. */
    file_id = (int64_t)H5Fcreate(udata->input.name, udata->input.flags, udata->input.fcpl_id, udata->input.fapl_id);

    /* Send response to client */
    HG_Handler_start_output(udata->hg_handle, &file_id);

    free(udata);
    udata = NULL;

    return;
} /* end nvol_file_create_handler_cb() */


static int
nvol_file_create_handler(hg_handle_t hg_handle)
{
    AXE_task_t task;
    file_create_ud_t *udata = NULL;

    if(NULL == (udata = (file_create_ud_t *)malloc(sizeof(file_create_ud_t))))
        ERROR("can't allocate input struct for decoding");

    if(HG_FAIL == HG_Handler_get_input(hg_handle, &udata->input))
        ERROR("can't get input parameters");

    assert(engine);

    udata->hg_handle = hg_handle;
    if(AXE_SUCCEED != AXEgenerate_task_id(engine, &task))
        ERROR("can't generate task id");
    if(AXE_SUCCEED != AXEcreate_task(engine, task, 0, NULL, 0, NULL, nvol_file_create_handler_cb, udata, NULL))
        ERROR("can't insert task into async engine");
    if(AXE_SUCCEED != AXEfinish(engine, task))
        ERROR("can't finish task");

    return(HG_SUCCESS);

error:
    if(udata)
        free(udata);

    return(HG_FAIL);
} /* end nvol_file_create_handler() */


static void
nvol_file_open_handler_cb(AXE_engine_t axe_engine, size_t num_necessary_parents,
    AXE_task_t necessary_parents[], size_t num_sufficient_parents,
    AXE_task_t sufficient_parents[], void *_udata)
{
    file_open_ud_t *udata = (file_open_ud_t *)_udata;
    file_open_out_t output;

    /* Open file */
    if((output.file_id = (int64_t)H5Fopen(udata->input.name, udata->input.flags, udata->input.fapl_id)) < 0)
        goto error;

    /* Get file creation property list */
    if((output.fcpl_id = H5Fget_create_plist((hid_t)output.file_id)) < 0) {
        (void)H5Fclose((hid_t)output.file_id);
        goto error;
    } /* end if */

    /* Error block - set output to failure values */
    if(0) {
error:
        output.file_id = -1;
        output.fcpl_id = NVOL_DEF_ID;
    } /* end if */

    /* Send response to client */
    HG_Handler_start_output(udata->hg_handle, &output);

    free(udata);
    udata = NULL;

    return;
} /* end nvol_file_open_handler_cb() */


static int
nvol_file_open_handler(hg_handle_t hg_handle)
{
    AXE_task_t task;
    file_open_ud_t *udata = NULL;

    if(NULL == (udata = (file_open_ud_t *)malloc(sizeof(file_open_ud_t))))
        ERROR("can't allocate input struct for decoding");

    if(HG_FAIL == HG_Handler_get_input(hg_handle, &udata->input))
        ERROR("can't get input parameters");

    assert(engine);

    udata->hg_handle = hg_handle;
    if(AXE_SUCCEED != AXEgenerate_task_id(engine, &task))
        ERROR("can't generate task id");
    if(AXE_SUCCEED != AXEcreate_task(engine, task, 0, NULL, 0, NULL, nvol_file_open_handler_cb, udata, NULL))
        ERROR("can't insert task into async engine");
    if(AXE_SUCCEED != AXEfinish(engine, task))
        ERROR("can't finish task");

    return(HG_SUCCESS);

error:
    if(udata)
        free(udata);

    return(HG_FAIL);
} /* end nvol_file_open_handler() */


static void
nvol_file_close_handler_cb(AXE_engine_t axe_engine,
    size_t num_necessary_parents, AXE_task_t necessary_parents[],
    size_t num_sufficient_parents, AXE_task_t sufficient_parents[],
    void *_udata)
{
    obj_id_ud_t *udata = (obj_id_ud_t *)_udata;
    int32_t ret;

    /* Close file.  No need to check for errors because we wouldn't do anything
     * different - just forward result to client. */
    ret = (int32_t)H5Fclose((hid_t)udata->obj_id);

    /* Send response to client */
    HG_Handler_start_output(udata->hg_handle, &ret);

    free(udata);
    udata = NULL;

    return;
} /* end nvol_file_close_handler_cb() */


static int
nvol_file_close_handler(hg_handle_t hg_handle)
{
    AXE_task_t task;
    obj_id_ud_t *udata = NULL;

    if(NULL == (udata = (obj_id_ud_t *)malloc(sizeof(obj_id_ud_t))))
        ERROR("can't allocate input struct for decoding");

    if(HG_FAIL == HG_Handler_get_input(hg_handle, &udata->obj_id))
        ERROR("can't get input parameters");

    assert(engine);

    udata->hg_handle = hg_handle;
    if(AXE_SUCCEED != AXEgenerate_task_id(engine, &task))
        ERROR("can't generate task id");
    if(AXE_SUCCEED != AXEcreate_task(engine, task, 0, NULL, 0, NULL, nvol_file_close_handler_cb, udata, NULL))
        ERROR("can't insert task into async engine");
    if(AXE_SUCCEED != AXEfinish(engine, task))
        ERROR("can't finish task");

    return(HG_SUCCESS);

error:
    if(udata)
        free(udata);

    return(HG_FAIL);
} /* end nvol_file_close_handler() */


static void
nvol_group_create_handler_cb(AXE_engine_t axe_engine,
    size_t num_necessary_parents, AXE_task_t necessary_parents[],
    size_t num_sufficient_parents, AXE_task_t sufficient_parents[],
    void *_udata)
{
    group_create_ud_t *udata = (group_create_ud_t *)_udata;
    int64_t grp_id;

    /* Create group.  No need to check for errors because we wouldn't do
     * anything different - just forward result to client. */
    grp_id = (int64_t)H5Gcreate2((hid_t)udata->input.loc_id, udata->input.name, udata->input.lcpl_id, udata->input.gcpl_id, udata->input.gapl_id);

    /* Send response to client */
    HG_Handler_start_output(udata->hg_handle, &grp_id);

    free(udata);
    udata = NULL;

    return;
} /* end nvol_group_create_handler_cb() */


static int
nvol_group_create_handler(hg_handle_t hg_handle)
{
    AXE_task_t task;
    group_create_ud_t *udata = NULL;

    if(NULL == (udata = (group_create_ud_t *)malloc(sizeof(group_create_ud_t))))
        ERROR("can't allocate input struct for decoding");

    if(HG_FAIL == HG_Handler_get_input(hg_handle, &udata->input))
        ERROR("can't get input parameters");

    assert(engine);

    udata->hg_handle = hg_handle;
    if(AXE_SUCCEED != AXEgenerate_task_id(engine, &task))
        ERROR("can't generate task id");
    if(AXE_SUCCEED != AXEcreate_task(engine, task, 0, NULL, 0, NULL, nvol_group_create_handler_cb, udata, NULL))
        ERROR("can't insert task into async engine");
    if(AXE_SUCCEED != AXEfinish(engine, task))
        ERROR("can't finish task");

    return(HG_SUCCESS);

error:
    if(udata)
        free(udata);

    return(HG_FAIL);
} /* end nvol_group_create_handler() */


static void
nvol_group_open_handler_cb(AXE_engine_t axe_engine, size_t num_necessary_parents,
    AXE_task_t necessary_parents[], size_t num_sufficient_parents,
    AXE_task_t sufficient_parents[], void *_udata)
{
    group_open_ud_t *udata = (group_open_ud_t *)_udata;
    group_open_out_t output;

    /* Open group */
    if((output.grp_id = (int64_t)H5Gopen2((hid_t)udata->input.loc_id, udata->input.name, udata->input.gapl_id)) < 0)
        goto error;

    /* Get group creation property list */
    if((output.gcpl_id = H5Gget_create_plist((hid_t)output.grp_id)) < 0) {
        (void)H5Gclose((hid_t)output.grp_id);
        goto error;
    } /* end if */

    /* Error block - set output to failure values */
    if(0) {
error:
        output.grp_id = -1;
        output.gcpl_id = NVOL_DEF_ID;
    } /* end if */

    /* Send response to client */
    HG_Handler_start_output(udata->hg_handle, &output);

    free(udata);
    udata = NULL;

    return;
} /* end nvol_group_open_handler_cb() */


static int
nvol_group_open_handler(hg_handle_t hg_handle)
{
    AXE_task_t task;
    group_open_ud_t *udata = NULL;

    if(NULL == (udata = (group_open_ud_t *)malloc(sizeof(group_open_ud_t))))
        ERROR("can't allocate input struct for decoding");

    if(HG_FAIL == HG_Handler_get_input(hg_handle, &udata->input))
        ERROR("can't get input parameters");

    assert(engine);

    udata->hg_handle = hg_handle;
    if(AXE_SUCCEED != AXEgenerate_task_id(engine, &task))
        ERROR("can't generate task id");
    if(AXE_SUCCEED != AXEcreate_task(engine, task, 0, NULL, 0, NULL, nvol_group_open_handler_cb, udata, NULL))
        ERROR("can't insert task into async engine");
    if(AXE_SUCCEED != AXEfinish(engine, task))
        ERROR("can't finish task");

    return(HG_SUCCESS);

error:
    if(udata)
        free(udata);

    return(HG_FAIL);
} /* end nvol_group_open_handler() */


static void
nvol_group_close_handler_cb(AXE_engine_t axe_engine,
    size_t num_necessary_parents, AXE_task_t necessary_parents[],
    size_t num_sufficient_parents, AXE_task_t sufficient_parents[],
    void *_udata)
{
    obj_id_ud_t *udata = (obj_id_ud_t *)_udata;
    int32_t ret;

    /* Close group.  No need to check for errors because we wouldn't do anything
     * different - just forward result to client. */
    ret = (int32_t)H5Gclose((hid_t)udata->obj_id);

    /* Send response to client */
    HG_Handler_start_output(udata->hg_handle, &ret);

    free(udata);
    udata = NULL;

    return;
} /* end nvol_group_close_handler_cb() */


static int
nvol_group_close_handler(hg_handle_t hg_handle)
{
    AXE_task_t task;
    obj_id_ud_t *udata = NULL;

    if(NULL == (udata = (obj_id_ud_t *)malloc(sizeof(obj_id_ud_t))))
        ERROR("can't allocate input struct for decoding");

    if(HG_FAIL == HG_Handler_get_input(hg_handle, &udata->obj_id))
        ERROR("can't get input parameters");

    assert(engine);

    udata->hg_handle = hg_handle;
    if(AXE_SUCCEED != AXEgenerate_task_id(engine, &task))
        ERROR("can't generate task id");
    if(AXE_SUCCEED != AXEcreate_task(engine, task, 0, NULL, 0, NULL, nvol_group_close_handler_cb, udata, NULL))
        ERROR("can't insert task into async engine");
    if(AXE_SUCCEED != AXEfinish(engine, task))
        ERROR("can't finish task");

    return(HG_SUCCESS);

error:
    if(udata)
        free(udata);

    return(HG_FAIL);
} /* end nvol_group_close_handler() */


static void
nvol_dataset_create_handler_cb(AXE_engine_t axe_engine,
    size_t num_necessary_parents, AXE_task_t necessary_parents[],
    size_t num_sufficient_parents, AXE_task_t sufficient_parents[],
    void *_udata)
{
    dset_create_ud_t *udata = (dset_create_ud_t *)_udata;
    int64_t dset_id;

    /* Create dataset.  No need to check for errors because we wouldn't do
     * anything different - just forward result to client. */
    dset_id = (int64_t)H5Dcreate2((hid_t)udata->input.loc_id, udata->input.name, udata->input.type_id, udata->input.space_id, udata->input.lcpl_id, udata->input.dcpl_id, udata->input.dapl_id);

    /* Send response to client */
    HG_Handler_start_output(udata->hg_handle, &dset_id);

    free(udata);
    udata = NULL;

    return;
} /* end nvol_dataset_create_handler_cb() */


static int
nvol_dataset_create_handler(hg_handle_t hg_handle)
{
    AXE_task_t task;
    dset_create_ud_t *udata = NULL;

    if(NULL == (udata = (dset_create_ud_t *)malloc(sizeof(dset_create_ud_t))))
        ERROR("can't allocate input struct for decoding");

    if(HG_FAIL == HG_Handler_get_input(hg_handle, &udata->input))
        ERROR("can't get input parameters");

    assert(engine);

    udata->hg_handle = hg_handle;
    if(AXE_SUCCEED != AXEgenerate_task_id(engine, &task))
        ERROR("can't generate task id");
    if(AXE_SUCCEED != AXEcreate_task(engine, task, 0, NULL, 0, NULL, nvol_dataset_create_handler_cb, udata, NULL))
        ERROR("can't insert task into async engine");
    if(AXE_SUCCEED != AXEfinish(engine, task))
        ERROR("can't finish task");

    return(HG_SUCCESS);

error:
    if(udata)
        free(udata);

    return(HG_FAIL);
} /* end nvol_dataset_create_handler() */


static void
nvol_dataset_open_handler_cb(AXE_engine_t axe_engine,
    size_t num_necessary_parents, AXE_task_t necessary_parents[],
    size_t num_sufficient_parents, AXE_task_t sufficient_parents[],
    void *_udata)
{
    dset_open_ud_t *udata = (dset_open_ud_t *)_udata;
    dset_open_out_t output;

    /* Open dataset */
    if((output.dset_id = (int64_t)H5Dopen2((hid_t)udata->input.loc_id, udata->input.name, udata->input.dapl_id)) < 0)
        goto error;

    /* Get dataset creation property list */
    if((output.dcpl_id = H5Dget_create_plist((hid_t)output.dset_id)) < 0) {
        (void)H5Dclose((hid_t)output.dset_id);
        goto error;
    } /* end if */

    /* Get dataset type */
    if((output.type_id = H5Dget_type((hid_t)output.dset_id)) < 0) {
        (void)H5Dclose((hid_t)output.dset_id);
        goto error;
    } /* end if */

    /* Get dataset space */
    if((output.space_id = H5Dget_space((hid_t)output.dset_id)) < 0) {
        (void)H5Dclose((hid_t)output.dset_id);
        goto error;
    } /* end if */

    /* Error block - set output to failure values */
    if(0) {
error:
        output.dset_id = -1;
        output.dcpl_id = NVOL_DEF_ID;
        output.type_id = NVOL_DEF_ID;
        output.space_id = NVOL_DEF_ID;
    } /* end if */

    /* Send response to client */
    HG_Handler_start_output(udata->hg_handle, &output);

    free(udata);
    udata = NULL;

    return;
} /* end nvol_dataset_open_handler_cb() */


static int
nvol_dataset_open_handler(hg_handle_t hg_handle)
{
    AXE_task_t task;
    dset_open_ud_t *udata = NULL;

    if(NULL == (udata = (dset_open_ud_t *)malloc(sizeof(dset_open_ud_t))))
        ERROR("can't allocate input struct for decoding");

    if(HG_FAIL == HG_Handler_get_input(hg_handle, &udata->input))
        ERROR("can't get input parameters");

    assert(engine);

    udata->hg_handle = hg_handle;
    if(AXE_SUCCEED != AXEgenerate_task_id(engine, &task))
        ERROR("can't generate task id");
    if(AXE_SUCCEED != AXEcreate_task(engine, task, 0, NULL, 0, NULL, nvol_dataset_open_handler_cb, udata, NULL))
        ERROR("can't insert task into async engine");
    if(AXE_SUCCEED != AXEfinish(engine, task))
        ERROR("can't finish task");

    return(HG_SUCCESS);

error:
    if(udata)
        free(udata);

    return(HG_FAIL);
} /* end nvol_dataset_open_handler() */


static void
nvol_dataset_read_handler_cb(AXE_engine_t axe_engine,
    size_t num_necessary_parents, AXE_task_t necessary_parents[],
    size_t num_sufficient_parents, AXE_task_t sufficient_parents[],
    void *_udata)
{
    dset_io_ud_t *udata = (dset_io_ud_t *)_udata;
    hid_t file_space_id = -1;
    hid_t tmp_space_id = H5S_ALL;
    hssize_t extent_npoints;
    hssize_t select_npoints;
    hsize_t space_dim;
    hg_bulk_block_t bulk_block_handle;
    hg_bulk_request_t bulk_request;
    size_t size;
    void *buf;
    na_addr_t dest;
    int32_t ret;

    /* Retrieve destination address */
    dest = HG_Handler_get_addr(udata->hg_handle);

    /* Retrieve space id */
    file_space_id = udata->input.file_space_id;
    if(file_space_id == NVOL_DEF_ID)
        file_space_id = H5S_ALL;

    /* Create dataspace to use for local buffer to read into. Must be
     * contiguous.  Use H5S_ALL if file_space_id is H5S_ALL or has all elements
     * selected.  Create 1-d dataspace otherwise.  Note tmp_space_id is
     * initialized to H5S_ALL. */
    if(file_space_id != H5S_ALL) {
        if((extent_npoints = H5Sget_simple_extent_npoints(file_space_id)) < 0)
            goto error;
        if((select_npoints = H5Sget_select_npoints(file_space_id)) < 0)
            goto error;
        assert(extent_npoints >= select_npoints);
        if(extent_npoints != select_npoints) {
            space_dim = (hsize_t)select_npoints;
            if((tmp_space_id = H5Screate_simple(1, &space_dim, NULL)) < 0)
                goto error;
        } /* end if */
    } /* end else */

    /* Read bulk data here and wait for the data to be here */
    size = HG_Bulk_handle_get_size(udata->input.bulk_handle);
    if(NULL == (buf = malloc(size)))
        goto error;

    HG_Bulk_block_handle_create(buf, size, HG_BULK_READ_ONLY, &bulk_block_handle);

    /* Read data from file */
    if((ret = (int32_t)H5Dread((hid_t)udata->input.dset_id, udata->input.mem_type_id, tmp_space_id, file_space_id, udata->input.dxpl_id, buf)) < 0)
        goto error;

    /* Write bulk data here and wait for the data to be there  */
    if(HG_SUCCESS != HG_Bulk_write_all(dest, udata->input.bulk_handle, bulk_block_handle, &bulk_request))
        goto error;
    /* wait for it to complete */
    if(HG_SUCCESS != HG_Bulk_wait(bulk_request, HG_BULK_MAX_IDLE_TIME, HG_BULK_STATUS_IGNORE))
        goto error;

    /* Free the bulk block handle */
    if(HG_SUCCESS != HG_Bulk_block_handle_free(bulk_block_handle))
        goto error;

    /* Error block - set output to failure values */
    if(0) {
error:
        ret = -1;
    } /* end if */

    /* Send response to client */
    HG_Handler_start_output(udata->hg_handle, &ret);

    if(tmp_space_id != H5S_ALL)
        (void)H5Sclose(tmp_space_id);
    free(udata);
    udata = NULL;

    return;
} /* end nvol_dataset_read_handler_cb() */


static int
nvol_dataset_read_handler(hg_handle_t hg_handle)
{
    AXE_task_t task;
    dset_io_ud_t *udata = NULL;

    if(NULL == (udata = (dset_io_ud_t *)malloc(sizeof(dset_io_ud_t))))
        ERROR("can't allocate input struct for decoding");

    if(HG_FAIL == HG_Handler_get_input(hg_handle, &udata->input))
        ERROR("can't get input parameters");

    assert(engine);

    udata->hg_handle = hg_handle;
    if(AXE_SUCCEED != AXEgenerate_task_id(engine, &task))
        ERROR("can't generate task id");
    if(AXE_SUCCEED != AXEcreate_task(engine, task, 0, NULL, 0, NULL, nvol_dataset_read_handler_cb, udata, NULL))
        ERROR("can't insert task into async engine");
    if(AXE_SUCCEED != AXEfinish(engine, task))
        ERROR("can't finish task");

    return(HG_SUCCESS);

error:
    if(udata)
        free(udata);

    return(HG_FAIL);
} /* end nvol_dataset_read_handler() */


static void
nvol_dataset_write_handler_cb(AXE_engine_t axe_engine,
    size_t num_necessary_parents, AXE_task_t necessary_parents[],
    size_t num_sufficient_parents, AXE_task_t sufficient_parents[],
    void *_udata)
{
    dset_io_ud_t *udata = (dset_io_ud_t *)_udata;
    hid_t file_space_id = -1;
    hid_t tmp_space_id = H5S_ALL;
    hssize_t extent_npoints;
    hssize_t select_npoints;
    hsize_t space_dim;
    hg_bulk_block_t bulk_block_handle;
    hg_bulk_request_t bulk_request;
    size_t size;
    void *buf;
    na_addr_t src;
    int32_t ret;

    /* Retrieve source address */
    src = HG_Handler_get_addr(udata->hg_handle);

    /* Retrieve space id */
    file_space_id = udata->input.file_space_id;
    if(file_space_id == NVOL_DEF_ID)
        file_space_id = H5S_ALL;

    /* Create dataspace to use for local buffer to write from. Must be
     * contiguous.  Use H5S_ALL if file_space_id is H5S_ALL or has all elements
     * selected.  Create 1-d dataspace otherwise.  Note tmp_space_id is
     * initialized to H5S_ALL. */
    if(file_space_id != H5S_ALL) {
        if((extent_npoints = H5Sget_simple_extent_npoints(file_space_id)) < 0)
            goto error;
        if((select_npoints = H5Sget_select_npoints(file_space_id)) < 0)
            goto error;
        assert(extent_npoints >= select_npoints);
        if(extent_npoints != select_npoints) {
            space_dim = (hsize_t)select_npoints;
            if((tmp_space_id = H5Screate_simple(1, &space_dim, NULL)) < 0)
                goto error;
        } /* end if */
    } /* end else */

    /* Read bulk data here and wait for the data to be here */
    size = HG_Bulk_handle_get_size(udata->input.bulk_handle);
    if(NULL == (buf = malloc(size)))
        goto error;

    HG_Bulk_block_handle_create(buf, size, HG_BULK_READWRITE, &bulk_block_handle);

    /* Write bulk data here and wait for the data to be there  */
    if(HG_SUCCESS != HG_Bulk_read_all(src, udata->input.bulk_handle, bulk_block_handle, &bulk_request)) {
        (void)HG_Bulk_block_handle_free(bulk_block_handle);
        goto error;
    } /* end if */
    if(HG_SUCCESS != HG_Bulk_wait(bulk_request, HG_BULK_MAX_IDLE_TIME, HG_BULK_STATUS_IGNORE)) {
        (void)HG_Bulk_block_handle_free(bulk_block_handle);
        goto error;
    } /* end if */

    /* Free the bulk block handle */
    if(HG_SUCCESS != HG_Bulk_block_handle_free(bulk_block_handle))
        goto error;

    /* Write data to file */
    if((ret = (int32_t)H5Dwrite((hid_t)udata->input.dset_id, udata->input.mem_type_id, tmp_space_id, file_space_id, udata->input.dxpl_id, buf)) < 0)
        goto error;

    /* Error block - set output to failure values */
    if(0) {
error:
        ret = -1;
    } /* end if */

    /* Send response to client */
    HG_Handler_start_output(udata->hg_handle, &ret);

    if(tmp_space_id != H5S_ALL)
        (void)H5Sclose(tmp_space_id);
    free(udata);
    udata = NULL;

    return;
} /* end nvol_dataset_write_handler_cb() */


static int
nvol_dataset_write_handler(hg_handle_t hg_handle)
{
    AXE_task_t task;
    dset_io_ud_t *udata = NULL;

    if(NULL == (udata = (dset_io_ud_t *)malloc(sizeof(dset_io_ud_t))))
        ERROR("can't allocate input struct for decoding");

    if(HG_FAIL == HG_Handler_get_input(hg_handle, &udata->input))
        ERROR("can't get input parameters");

    assert(engine);

    udata->hg_handle = hg_handle;
    if(AXE_SUCCEED != AXEgenerate_task_id(engine, &task))
        ERROR("can't generate task id");
    if(AXE_SUCCEED != AXEcreate_task(engine, task, 0, NULL, 0, NULL, nvol_dataset_write_handler_cb, udata, NULL))
        ERROR("can't insert task into async engine");
    if(AXE_SUCCEED != AXEfinish(engine, task))
        ERROR("can't finish task");

    return(HG_SUCCESS);

error:
    if(udata)
        free(udata);

    return(HG_FAIL);
} /* end nvol_dataset_write_handler() */


static void
nvol_dataset_close_handler_cb(AXE_engine_t axe_engine,
    size_t num_necessary_parents, AXE_task_t necessary_parents[],
    size_t num_sufficient_parents, AXE_task_t sufficient_parents[],
    void *_udata)
{
    obj_id_ud_t *udata = (obj_id_ud_t *)_udata;
    int32_t ret;

    /* Close dataset.  No need to check for errors because we wouldn't do
     * anything different - just forward result to client. */
    ret = (int32_t)H5Dclose((hid_t)udata->obj_id);

    /* Send response to client */
    HG_Handler_start_output(udata->hg_handle, &ret);

    free(udata);
    udata = NULL;

    return;
} /* end nvol_dataset_close_handler_cb() */


static int
nvol_dataset_close_handler(hg_handle_t hg_handle)
{
    AXE_task_t task;
    obj_id_ud_t *udata = NULL;

    if(NULL == (udata = (obj_id_ud_t *)malloc(sizeof(obj_id_ud_t))))
        ERROR("can't allocate input struct for decoding");

    if(HG_FAIL == HG_Handler_get_input(hg_handle, &udata->obj_id))
        ERROR("can't get input parameters");

    assert(engine);

    udata->hg_handle = hg_handle;
    if(AXE_SUCCEED != AXEgenerate_task_id(engine, &task))
        ERROR("can't generate task id");
    if(AXE_SUCCEED != AXEcreate_task(engine, task, 0, NULL, 0, NULL, nvol_dataset_close_handler_cb, udata, NULL))
        ERROR("can't insert task into async engine");
    if(AXE_SUCCEED != AXEfinish(engine, task))
        ERROR("can't finish task");

    return(HG_SUCCESS);

error:
    if(udata)
        free(udata);

    return(HG_FAIL);
} /* end nvol_dataset_close_handler() */


int
hdf5_network_vol_start_handler(na_class_t *network_class)
{
    AXE_engine_attr_t engine_attr;
    herr_t ret_value = SUCCEED;

    if(!network_class)
        ERROR("no network class provided");

    /* initialize the network class */
    if(HG_SUCCESS != HG_Handler_init(network_class))
        ERROR("Could not initialize function shipper\n");
    if(HG_SUCCESS != HG_Bulk_init(network_class))
        ERROR("Could not initialize bulk data shipper\n");

    /* Register function and encoding/decoding functions */
    MERCURY_HANDLER_REGISTER("file_create", nvol_file_create_handler, 
                             file_create_in_t, servid_t);
    MERCURY_HANDLER_REGISTER("file_open", nvol_file_open_handler, 
                             file_open_in_t, file_open_out_t);
    /*MERCURY_HANDLER_REGISTER("file_flush", H5VL_iod_server_file_flush, 
                             file_flush_in_t, ret_t);*/
    MERCURY_HANDLER_REGISTER("file_close", nvol_file_close_handler, 
                             servid_t, ret_t);

    /*MERCURY_HANDLER_REGISTER("attr_create", H5VL_iod_server_attr_create, 
                             attr_create_in_t, attr_create_out_t);
    MERCURY_HANDLER_REGISTER("attr_open", H5VL_iod_server_attr_open, 
                             attr_open_in_t, attr_open_out_t);
    MERCURY_HANDLER_REGISTER("attr_read", H5VL_iod_server_attr_read, 
                             attr_io_in_t, ret_t);
    MERCURY_HANDLER_REGISTER("attr_write", H5VL_iod_server_attr_write, 
                             attr_io_in_t, ret_t);
    MERCURY_HANDLER_REGISTER("attr_exists", H5VL_iod_server_attr_exists, 
                             attr_op_in_t, htri_t);
    MERCURY_HANDLER_REGISTER("attr_rename", H5VL_iod_server_attr_rename, 
                             attr_rename_in_t, ret_t);
    MERCURY_HANDLER_REGISTER("attr_remove", H5VL_iod_server_attr_remove, 
                             attr_op_in_t, ret_t);
    MERCURY_HANDLER_REGISTER("attr_close", H5VL_iod_server_attr_close, 
                             attr_close_in_t, ret_t);*/

    MERCURY_HANDLER_REGISTER("group_create", nvol_group_create_handler, 
                             group_create_in_t, servid_t);
    MERCURY_HANDLER_REGISTER("group_open", nvol_group_open_handler, 
                             group_open_in_t, group_open_out_t);
    MERCURY_HANDLER_REGISTER("group_close", nvol_group_close_handler, 
                             servid_t, ret_t);

    MERCURY_HANDLER_REGISTER("dset_create", nvol_dataset_create_handler, 
                             dset_create_in_t, servid_t);
    MERCURY_HANDLER_REGISTER("dset_open", nvol_dataset_open_handler, 
                             dset_open_in_t, dset_open_out_t);
    MERCURY_HANDLER_REGISTER("dset_read", nvol_dataset_read_handler, 
                             dset_io_in_t, ret_t);
    MERCURY_HANDLER_REGISTER("dset_write", nvol_dataset_write_handler, 
                             dset_io_in_t, ret_t);
    /*MERCURY_HANDLER_REGISTER("dset_set_extent", H5VL_iod_server_dset_set_extent, 
                             dset_set_extent_in_t, ret_t);*/
    MERCURY_HANDLER_REGISTER("dset_close", nvol_dataset_close_handler, 
                             servid_t, ret_t);

    /*MERCURY_HANDLER_REGISTER("dtype_commit", H5VL_iod_server_dtype_commit, 
                             dtype_commit_in_t, dtype_commit_out_t);
    MERCURY_HANDLER_REGISTER("dtype_open", H5VL_iod_server_dtype_open, 
                             dtype_open_in_t, dtype_open_out_t);
    MERCURY_HANDLER_REGISTER("dtype_close", H5VL_iod_server_dtype_close, 
                             dtype_close_in_t, ret_t);

    MERCURY_HANDLER_REGISTER("link_create", H5VL_iod_server_link_create, 
                             link_create_in_t, ret_t);
    MERCURY_HANDLER_REGISTER("link_move", H5VL_iod_server_link_move, 
                             link_move_in_t, ret_t);
    MERCURY_HANDLER_REGISTER("link_exists", H5VL_iod_server_link_exists, 
                             link_op_in_t, htri_t);
    MERCURY_HANDLER_REGISTER("link_iterate", H5VL_iod_server_link_iterate, 
                             link_op_in_t, ret_t);
    MERCURY_HANDLER_REGISTER("link_remove", H5VL_iod_server_link_remove, 
                             link_op_in_t, ret_t);

    MERCURY_HANDLER_REGISTER("object_open", H5VL_iod_server_object_open, 
                             object_op_in_t, object_open_out_t);
    MERCURY_HANDLER_REGISTER("object_copy", H5VL_iod_server_object_copy, 
                             object_copy_in_t, ret_t);
    MERCURY_HANDLER_REGISTER("object_exists", H5VL_iod_server_object_exists, 
                             object_op_in_t, htri_t);
    MERCURY_HANDLER_REGISTER("object_visit", H5VL_iod_server_object_visit, 
                             object_op_in_t, ret_t);
    MERCURY_HANDLER_REGISTER("set_comment", H5VL_iod_server_object_set_comment, 
                             object_set_comment_in_t, ret_t);
    MERCURY_HANDLER_REGISTER("get_comment", H5VL_iod_server_object_get_comment, 
                             object_get_comment_in_t, object_get_comment_out_t);

    MERCURY_HANDLER_REGISTER("cancel_op", H5VL_iod_server_cancel_op, uint64_t, uint8_t);*/

    /* Initialize engine attribute */
    if(AXEengine_attr_init(&engine_attr) != AXE_SUCCEED)
        ERROR("unable to initialize AXE engine attribute");

    /* Set number of threads in AXE engine */
    if(AXEset_num_threads(&engine_attr, 4) != AXE_SUCCEED)
        ERROR("unable to set number of threads in AXE engine");

    /* Create AXE engine */
    if(AXEcreate_engine(&engine, &engine_attr) != AXE_SUCCEED)
        ERROR("unable to create AXE engine");

    /* Loop to receive requests from clients */
    do {
        /* Receive new function calls */
        if(HG_SUCCESS != HG_Handler_process(HG_HANDLER_MAX_IDLE_TIME, HG_STATUS_IGNORE))
            ERROR("function shipper failed to handle request");
    } while(!nvol_shutdown);

    if(AXE_SUCCEED != AXEterminate_engine(engine, TRUE))
        ERROR("unable to terminate AXE engine");

    if(AXEengine_attr_destroy(&engine_attr) != AXE_SUCCEED)
        ERROR("unable to destroy engine attribute");

    if(HG_SUCCESS != HG_Bulk_finalize())
        ERROR("unable to finalize bulk data shipper");
    if(HG_SUCCESS != HG_Handler_finalize())
        ERROR("unable to finalize function handler");

    return ret_value;

error:
    return FAIL;
}

